/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.grpc;

import static de.ozgcloud.vorgang.common.grpc.FormDataMappingUtils.*;
import static de.ozgcloud.vorgang.common.grpc.GrpcSubFormTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Spy;

import de.ozgcloud.vorgang.vorgang.GrpcFormData;
import de.ozgcloud.vorgang.vorgang.GrpcFormField;
import de.ozgcloud.vorgang.vorgang.GrpcSubForm;

class GrpcFormDataMapperTest {

	@Spy
	@InjectMocks
	private GrpcFormDataMapper mapper = Mappers.getMapper(GrpcFormDataMapper.class);

	@Nested
	class TestMapToAllFields {
		@Test
		void shouldMapFieldForStringValue() {
			mapper.mapToAllFields(Map.of(FIELD_NAME, FIELD_VALUE));

			verify(mapper).mapToField(FIELD_NAME, FIELD_VALUE);
		}

		@Test
		void shouldMapFieldForValueMap() {
			Map<String, Object> valueMap = Map.of(VALUE_KEY, FIELD_VALUE);

			mapper.mapToAllFields(Map.of(FIELD_NAME, valueMap));

			verify(mapper).mapToField(FIELD_NAME, valueMap);
		}

		@Test
		void shouldNOTMapFieldForSubForm() {
			Map<String, Object> subForm = Map.of(SUBFORM_TITLE, Map.of(FIELD_NAME, FIELD_VALUE));

			mapper.mapToAllFields(Map.of(FIELD_NAME, subForm));

			verify(mapper, never()).mapToField(any(), any());
		}

		@Test
		void souldNOTMapFieldForSubFormValue() {
			Map<String, Object> subForm = Map.of(SUBFORM_TITLE, Map.of(FIELD_NAME, FIELD_VALUE));

			mapper.mapToAllFields(Map.of(FIELD_NAME, Map.of(VALUE_KEY, subForm)));

			verify(mapper, never()).mapToField(any(), any());
		}

		@Nested
		class MapToField {

			private Map<String, Object> valueContainer = Map.of(VALUE_KEY, FIELD_VALUE, LABEL_KEY, FIELD_LABEL);

			@Test
			void shouldHaveName() {
				var grpcField = mapper.mapToField(FIELD_NAME, valueContainer);

				assertThat(grpcField.getName()).isEqualTo(FIELD_NAME);
			}

			@Test
			void shouldHaveValue() {
				var grpcField = mapper.mapToField(FIELD_NAME, valueContainer);

				assertThat(grpcField.getValue()).isEqualTo(FIELD_VALUE);
			}

			@Test
			void shouldHaveLabel() {
				var grpcField = mapper.mapToField(FIELD_NAME, valueContainer);

				assertThat(grpcField.getLabel()).isEqualTo(FIELD_LABEL);
			}

			@Test
			void shouldHaveNumeric() {
				var grpcField = mapper.mapToField(FIELD_NAME, Map.of(VALUE_KEY, 42, LABEL_KEY, FIELD_LABEL));

				assertThat(grpcField.getValue()).isEqualTo("42");
			}
		}
	}

	@Nested
	class TestSimpleValueMapping {

		@Test
		void shouldHaveField() {
			GrpcFormData formData = mapper.mapToFormData(Map.of("key", "value"));

			assertThat(formData.getFieldList()).hasSize(1);
		}

		@Test
		void shouldNotHaveSubForms() {

			GrpcFormData formData = mapper.mapToFormData(Map.of("key", "value"));

			assertThat(formData.getFormList()).isEmpty();
		}

		@Test
		void shouldHaveNameField() {

			GrpcFormData formData = mapper.mapToFormData(Map.of("key", "value"));

			assertThat(formData.getField(0).getName()).isEqualTo("key");
			assertThat(formData.getField(0).getValue()).isEqualTo("value");
		}
	}

	@Nested
	class TestSubFormMapping {

		@Test
		void shouldHaveSubForm() {

			GrpcFormData formData = mapper.mapToFormData(Map.of("key", Map.of("subKey", "value")));

			assertThat(formData.getFormList()).hasSize(1);
		}

		@Test
		void shouldNotHaveFields() {

			GrpcFormData formData = mapper.mapToFormData(Map.of("key", Map.of("subKey", "value")));

			assertThat(formData.getFieldList()).isEmpty();
		}

		@Test
		void shouldHaveSubFormField() {
			GrpcFormData formData = mapper.mapToFormData(Map.of("key", Map.of("subKey", "value")));

			assertThat(formData.getForm(0).getFieldList()).hasSize(1);
			assertThat(formData.getForm(0).getTitle()).isEqualTo("key");
			assertThat(formData.getForm(0).getField(0).getName()).isEqualTo("subKey");
			assertThat(formData.getForm(0).getField(0).getValue()).isEqualTo("value");
		}

		@Test
		void shouldHaveNumericSubFormField() {
			GrpcFormData formData = mapper.mapToFormData(Map.of("key", Map.of("subKey", 42)));

			assertThat(formData.getForm(0).getFieldList()).hasSize(1);
			assertThat(formData.getForm(0).getField(0).getValue()).isEqualTo("42");
		}

	}

	@DisplayName("Mapped SubFormField")
	@Nested
	class TestSubFormFieldMapping {

		private final GrpcSubForm subForm = GrpcSubFormTestFactory.create();

		@Test
		void shouldContainField() {
			var mapped = mapper.mapSubFormFields(subForm.getFieldList());

			assertThat(mapped).containsKey(GrpcSubFormTestFactory.FIELD_NAME);
		}

		@Test
		void shouldContainValue() {
			var mapped = mapper.mapSubFormFields(subForm.getFieldList());

			assertThat(mapped).containsEntry(GrpcSubFormTestFactory.FIELD_NAME, GrpcSubFormTestFactory.FIELD);
		}

		@Nested
		class TestCombiner {
			@Test
			void missingInMap2ShouldContainValue() {
				Map<String, Object> map1 = Map.of("TEST", "VALUE1");
				Map<String, Object> map2 = Map.of();

				var combined = mapper.combiner(map1, map2);

				assertThat(combined).containsEntry("TEST", "VALUE1");
			}

			@Test
			void missingInMap1ShouldContainValue() {
				Map<String, Object> map1 = new HashMap<>();
				Map<String, Object> map2 = Map.of("TEST", "VALUE1");

				var combined = mapper.combiner(map1, map2);

				assertThat(combined).containsEntry("TEST", Map.of(VALUE_KEY, "VALUE1"));
			}

			@Test
			void inBothMapShouldCombineToList() {
				Map<String, Object> map1 = new HashMap<>(Map.of("TEST", "VALUE1"));
				Map<String, Object> map2 = Map.of("TEST", "VALUE2");

				var combined = mapper.combiner(map1, map2);

				assertThat(combined).containsEntry("TEST", Map.of(VALUE_KEY, List.of("VALUE1", "VALUE2")));
			}

			@Test
			void listInMap2ShouldStayAList() {
				Map<String, Object> map1 = Map.of("TEST", List.of("VALUE1", "VALUE2"));
				Map<String, Object> map2 = Map.of();

				var combined = mapper.combiner(map1, map2);

				assertThat(combined).containsEntry("TEST", List.of("VALUE1", "VALUE2"));
			}

			@Test
			void valueInMap1ShouldBeAddedToList() {
				Map<String, Object> map1 = new HashMap<>(Map.of("TEST", "VALUE1"));
				Map<String, Object> map2 = Map.of("TEST", List.of("VALUE2", "VALUE3"));

				var combined = mapper.combiner(map1, map2);

				assertThat(combined).containsEntry("TEST", Map.of(VALUE_KEY, List.of("VALUE1", "VALUE2", "VALUE3")));
			}

			@Test
			void valueInMap2ShouldBeAddedToList() {
				Map<String, Object> map1 = new HashMap<>(Map.of("TEST", List.of("VALUE1", "VALUE2")));
				Map<String, Object> map2 = Map.of("TEST", "VALUE3");

				var combined = mapper.combiner(map1, map2);

				assertThat(combined).containsEntry("TEST", Map.of(VALUE_KEY, List.of("VALUE1", "VALUE2", "VALUE3")));
			}

			@Test
			void listInMap1ShouldStayAList() {
				Map<String, Object> map1 = new HashMap<>();
				Map<String, Object> map2 = Map.of("TEST", List.of("VALUE1", "VALUE2"));

				var combined = mapper.combiner(map1, map2);

				assertThat(combined).containsEntry("TEST", Map.of(VALUE_KEY, List.of("VALUE1", "VALUE2")));
			}

			@Test
			void listInBothMapsShouldByCombined() {
				Map<String, Object> map1 = new HashMap<>(Map.of("TEST", List.of("VALUE1", "VALUE2")));
				Map<String, Object> map2 = Map.of("TEST", List.of("VALUE3", "VALUE4"));

				var combined = mapper.combiner(map1, map2);

				assertThat(combined).containsEntry("TEST", Map.of(VALUE_KEY, List.of("VALUE1", "VALUE2", "VALUE3", "VALUE4")));
			}
		}
	}

	@Nested
	class TestMapFormData {

		@Test
		void shouldHaveTitleAsKey() {
			var mapped = mapper.mapFormData(List.of(GrpcSubFormTestFactory.create()));

			assertThat(mapped).containsKey(TITLE);
		}

		@SuppressWarnings("unchecked")
		@Test
		void shouldHaveLabel() {
			var mapped = mapper.mapFormData(List.of(GrpcSubFormTestFactory.create()));

			assertThat((Map<String, Object>) mapped.get(TITLE)).containsEntry(LABEL_KEY, LABEL);
		}

		@Test
		void shouldHaveSubFormField() {
			var mapped = mapKontaktSubForm();

			assertThat(mapped).containsKey(FIELD_NAME);
		}

		@SuppressWarnings("unchecked")
		private Map<String, Object> mapKontaktSubForm() {
			var mapped = mapper.mapFormData(List.of(GrpcSubFormTestFactory.create()));

			var formMap = (Map<String, Object>) mapped.get(TITLE);
			return (Map<String, Object>) formMap.get(VALUE_KEY);
		}

		@SuppressWarnings("unchecked")
		@Test
		void shouldJoinIntoListInValue() {
			var mapped = mapper.mapFormData(List.of(GrpcSubFormTestFactory.create(), GrpcSubFormTestFactory.create()));

			var formList = (Map<String, Object>) mapped.get(TITLE);

			assertThat(formList).containsKey(VALUE_KEY);
			var value = (List<Map<String, Object>>) formList.get(VALUE_KEY);
			assertThat(value).hasSize(2);
		}
	}

	@Nested
	class TestMapStringListsToFields {

		@Test
		void emptyMapShouldNotBeMapped() {
			List<GrpcFormField> fields = mapper.mapStringListsToFields(Collections.emptyMap());

			assertThat(fields).isEmpty();
		}

		@Test
		void simpleValueShouldNotBeMapped() {
			List<GrpcFormField> fields = mapper.mapStringListsToFields(Map.of("a", "b"));

			assertThat(fields).isEmpty();
		}

		@Test
		void listObjectValuesShouldBeMapped() {
			List<GrpcFormField> fields = mapper.mapStringListsToFields(Map.of("key", List.of("value1", "value2")));

			assertThat(fields).hasSize(2);
			assertThat(fields.get(0).getName()).isEqualTo("key");
			assertThat(fields.get(0).getValue()).isEqualTo("value1");

			assertThat(fields.get(1).getName()).isEqualTo("key");
			assertThat(fields.get(1).getValue()).isEqualTo("value2");
		}

		@Test
		void shouldMapValueList() {
			var fields = mapper.mapStringListsToFields(Map.of(FIELD_NAME, Map.of(LABEL_KEY, FIELD_LABEL, VALUE_KEY, List.of(FIELD_VALUE))));

			assertThat(fields).hasSize(1).first().extracting(GrpcFormField::getValue).isEqualTo(FIELD_VALUE);
		}

		@Test
		void shouldAddLabel() {
			var fields = mapper.mapStringListsToFields(Map.of(FIELD_NAME, Map.of(LABEL_KEY, FIELD_LABEL, VALUE_KEY, List.of(FIELD_VALUE))));

			assertThat(fields).hasSize(1).first().extracting(GrpcFormField::getLabel).isEqualTo(FIELD_LABEL);
		}
	}

	@Nested
	class TestMapObjectListsToFields {

		@Captor
		private ArgumentCaptor<Entry<String, Object>> entryCaptor;

		@Test
		void simpleValueShouldNotBeMapped() {
			List<GrpcSubForm> fields = mapper.mapObjectListsToFields(Map.of("a", "b"));

			assertThat(fields).isEmpty();
		}

		@Test
		void listOfSimpleValueShouldNotBeMapped() {
			List<GrpcSubForm> fields = mapper.mapObjectListsToFields(Map.of("a", List.of("l1", "l2")));

			assertThat(fields).isEmpty();
		}

		@Test
		void listOfObjectsShouldBeMappedToSubForms() {
			List<GrpcSubForm> fields = mapper
					.mapObjectListsToFields(Map.of("key", List.of(Collections.emptyMap(), Collections.emptyMap())));

			assertThat(fields).hasSize(2);
		}

		@Test
		void listOfObjectsShouldMappedToSubFormTitle() {
			List<GrpcSubForm> fields = mapper.mapObjectListsToFields(Map.of("key", List.of(Collections.emptyMap(), Collections.emptyMap())));

			assertThat(fields.get(0).getTitle()).isEqualTo("key");
			assertThat(fields.get(1).getTitle()).isEqualTo("key");
		}

		@Test
		void listOfObjectsShouldBeMapped() {
			List<GrpcSubForm> fields = mapper.mapObjectListsToFields(Map.of("key", List.of(Map.of("a1", "a2"))));

			assertThat(fields.get(0).getFieldCount()).isEqualTo(1);
			assertThat(fields.get(0).getField(0).getName()).isEqualTo("a1");
			assertThat(fields.get(0).getField(0).getValue()).isEqualTo("a2");
		}

		@Test
		void shouldMapSubForm() {
			List<GrpcSubForm> subForms = mapper.mapObjectListsToFields(Map.of(SUBFORM_TITLE, List.of(SUBFORM)));

			assertThat(subForms).hasSize(1);

			assertThat(subForms.get(0).getField(0).getName()).isEqualTo(SUBFORM_FIELD_NAME);
			assertThat(subForms.get(0).getField(0).getValue()).isEqualTo(SUBFORM_FIELD_VALUE);
		}

		@Test
		void shouldMapListOfObjectsWithLabel() {
			List<GrpcSubForm> subForms = mapper.mapObjectListsToFields(Map.of(SUBFORM_TITLE, List.of(SUBFORM)));

			assertThat(subForms).hasSize(1);
			var subForm = subForms.get(0);

			assertThat(subForm.getLabel()).isEqualTo(SUBFORM_LABEL);
			assertThat(subForm.getField(0).getLabel()).isEqualTo(SUBFORM_FIELD_LABEL);
		}

		@Test
		void doubleNestedListObjectValuesShouldBeMapped() {
			GrpcFormData formData = mapper.mapToFormData(Map.of("key1", Map.of("key2", List.of("value1", "value2"))));

			assertThat(formData.getForm(0).getFieldCount()).isEqualTo(2);
		}

		@Test
		void multipleNestedListObjectValuesShouldBeMapped() {
			GrpcFormData formData = mapper.mapToFormData(Map.of("key1", Map.of("key2", Map.of("key3", List.of("value1", "value2")))));

			assertThat(formData.getForm(0).getSubForm(0).getFieldCount()).isEqualTo(2);
		}

		@Test
		void shouldCallMappingMethodForValueList() {
			mapper.mapObjectListsToFields(Map.of(FIELD_NAME, Map.of(LABEL_KEY, FIELD_LABEL, VALUE_KEY, List.of(SUBFORM))));

			verify(mapper).mapListObjectElementsToFormFields(entryCaptor.capture());
			assertThat(entryCaptor.getValue().getKey()).isEqualTo(FIELD_NAME);
			assertThat(entryCaptor.getValue().getValue()).isInstanceOf(Map.class);
		}

		@Test
		void shouldAddLabel() {
			var subForms = mapper.mapObjectListsToFields(Map.of(FIELD_NAME, Map.of(LABEL_KEY, FIELD_LABEL, VALUE_KEY, List.of(SUBFORM))));

			assertThat(subForms).hasSize(1).first().extracting(GrpcSubForm::getLabel).isEqualTo(SUBFORM_LABEL);
		}
	}

	@Nested
	class TestMapListOfMixedValuesInFormData {

		@Test
		void shouldMapListOfStrings() {
			GrpcFormData formData = mapper
					.mapToFormData(Map.of("key", List.of("value1", "value2", Map.of("internKey1", "internValue1"))));

			assertThat(formData.getFieldCount()).isEqualTo(2);
			assertThat(formData.getFormCount()).isEqualTo(1);
		}
	}

	@Nested
	class TestMapListObjectElementsToFormFields {
		@Test
		void shouldMapValueMap() {
			Map<String, Object> formData = Map.of(FIELD_NAME, Map.of(LABEL_KEY, FIELD_LABEL, VALUE_KEY, List.of(SUBFORM)));

			var result = mapper.mapListObjectElementsToFormFields(formData.entrySet().iterator().next());

			assertThat(result).hasSize(1).first().extracting(GrpcSubForm::getTitle).isEqualTo(FIELD_NAME);
		}

		@Test
		void shouldAddLabel() {
			Map<String, Object> formData = Map.of(FIELD_NAME, Map.of(LABEL_KEY, FIELD_LABEL, VALUE_KEY, List.of(SUBFORM)));

			var result = mapper.mapListObjectElementsToFormFields(formData.entrySet().iterator().next());

			assertThat(result).hasSize(1).first().extracting(GrpcSubForm::getLabel).isEqualTo(SUBFORM_LABEL);
		}
	}

	@Nested
	class TestBuildSubForm {

		@Test
		void shouldHaveLabel() {
			var grpc = mapper.buildSubForm(SUBFORM_TITLE, SUBFORM);

			assertThat(grpc.getLabel()).isEqualTo(SUBFORM_LABEL);
		}

	}
}
