/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.grpc;

import static de.ozgcloud.vorgang.common.grpc.GrpcObjectTestFactory.createAsMap;
import static de.ozgcloud.vorgang.common.grpc.GrpcPropertyTestFactory.*;
import static de.ozgcloud.vorgang.common.grpc.GrpcSubObjectTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.Spy;

import de.ozgcloud.common.binaryfile.FileId;
import de.ozgcloud.vorgang.common.GrpcObject;
import de.ozgcloud.vorgang.common.GrpcProperty;
import de.ozgcloud.vorgang.common.GrpcSubObject;

class GrpcObjectMapperTest {

	@Spy
	private final GrpcObjectMapper mapper = Mappers.getMapper(GrpcObjectMapper.class);

	@DisplayName("From map")
	@Nested
	class TestFromMap {

		@Test
		void shouldCallMapToAllProperties() {
			var sourceMap = createAsMap();

			mapper.fromMap(sourceMap);

			verify(mapper).mapToAllProperties(sourceMap);
		}

		@Test
		void shouldCallMapToSubObjects() {
			var sourceMap = createAsMap();

			mapper.fromMap(sourceMap);

			verify(mapper).mapToSubObjects(sourceMap);
		}

		@Test
		void shouldMapToGrpcObject() {
			var sourceMap = createAsMap();

			var grpcObject = mapper.fromMap(sourceMap);

			assertThat(grpcObject).isEqualTo(GrpcObjectTestFactory.create());
		}

		@DisplayName("should map type")
		@Nested
		class TestByType {

			@DisplayName("Long")
			@Test
			void shouldMapLongMapToGrpc() {
				var value = 42L;
				var expectedProperty = GrpcPropertyTestFactory.createWithValue(value);

				var grpcObject = mapFromMap(value);

				assertThat(grpcObject.getPropertyList()).hasSize(1).first().isEqualTo(expectedProperty);
			}

			@DisplayName("Integer")
			@Test
			void shouldMapIntegerMapToGrpc() {
				var expectedProperty = GrpcPropertyTestFactory.createWithValue(Integer.MAX_VALUE);

				var grpcObject = mapFromMap(Integer.MAX_VALUE);

				assertThat(grpcObject.getPropertyList()).hasSize(1).first().isEqualTo(expectedProperty);
			}

			@DisplayName("Boolean")
			@Test
			void shouldMapBooleanMapToGrpc() {
				var expectedProperty = GrpcPropertyTestFactory.createWithValue(Boolean.TRUE);

				var grpcObject = mapFromMap(Boolean.TRUE);

				assertThat(grpcObject.getPropertyList()).hasSize(1).first().isEqualTo(expectedProperty);
			}

			@DisplayName("LocalDate")
			@Test
			void shouldMapLocalDateMapToGrpc() {
				var date = "2022-12-11";
				var expectedProperty = GrpcPropertyTestFactory.createWithValue(date);

				var grpcObject = mapFromMap(LocalDate.parse(date));

				assertThat(grpcObject.getPropertyList()).hasSize(1).first().isEqualTo(expectedProperty);
			}

			@DisplayName("ZonedDateTime")
			@Test
			void shouldMapZonedDateTimeMapToGrpc() {
				var zonedDateTime = "2007-12-03T10:15:30+01:00";
				var expectedProperty = GrpcPropertyTestFactory.createWithValue(zonedDateTime);

				var grpcObject = mapFromMap(ZonedDateTime.parse(zonedDateTime));

				assertThat(grpcObject.getPropertyList()).hasSize(1).first().isEqualTo(expectedProperty);
			}

			@DisplayName("FileId")
			@Test
			void shouldMapFileIdMapToGrpc() {
				var fileId = FileId.createNew();
				var expectedProperty = GrpcPropertyTestFactory.createWithValue(fileId);

				var grpcObject = mapFromMap(fileId);

				assertThat(grpcObject.getPropertyList()).hasSize(1).first().isEqualTo(expectedProperty);
			}

			private GrpcObject mapFromMap(Object value) {
				return mapper.fromMap(Map.of(PROPERTY_NAME, value));
			}

		}
	}

	@DisplayName("Map all properties")
	@Nested
	class TestMapAllProperties {

		private Map<String, Object> propertyMap = Map.of(PROPERTY_NAME, PROPERTY_NAME_VALUE);

		@Test
		void shouldCallMapToProperty() {
			mapper.mapToAllProperties(propertyMap);

			verify(mapper, times(1)).mapToProperty(PROPERTY_NAME, PROPERTY_NAME_VALUE);
		}

		@Test
		void shouldMapToProperty() {
			var grpcProperties = mapper.mapToAllProperties(propertyMap);

			assertThat(grpcProperties).isEqualTo(List.of(GrpcPropertyTestFactory.create()));
		}

		@Test
		void shouldNotCallMapToPropertyOnMap() {
			mapper.mapToAllProperties(Map.of(PROPERTY_NAME, propertyMap));

			verify(mapper, never()).mapToProperty(any(), any());
		}
	}

	@DisplayName("Map property")
	@Nested
	class TestMapProperty {

		@DisplayName("should map null value of property to an empty string")
		@Test
		void shouldMapNullValueToEmptyString() {
			var grpcProperty = mapper.mapToProperty(PROPERTY_NAME, null);

			assertThat(grpcProperty.getValueList()).containsExactly(StringUtils.EMPTY);
		}

		@DisplayName("should map if value is a collection")
		@Test
		void shouldMapValueAsCollection() {
			var grpcProperty = mapper.mapToProperty(PROPERTY_NAME, Set.of(PROPERTY_NAME_VALUE, "another_value"));

			assertThat(grpcProperty.getValueList()).containsExactlyInAnyOrder(PROPERTY_NAME_VALUE, "another_value");
		}

		@DisplayName("should map if value is a collection containing null")
		@Test
		void shouldMapValueAsCollectionContainsNull() {
			var listContainsNull = Collections.singletonList(null);

			var grpcProperty = mapper.mapToProperty(PROPERTY_NAME, listContainsNull);

			assertThat(grpcProperty.getValueList()).containsExactly(StringUtils.EMPTY);
		}

		@DisplayName("should map if value is a list considering the order of the values")
		@Test
		void shouldMapValueAsListConsideringOrder() {
			var list = List.of("firstProperty", "secondProperty");

			var grpcProperty = mapper.mapToProperty(PROPERTY_NAME, list);

			assertThat(grpcProperty.getValueList()).containsExactly("firstProperty", "secondProperty");
		}

	}

	@DisplayName("Map subObject")
	@Nested
	class TestMapSubObject {

		@Test
		void shouldCallBuildSubObject() {
			var propertyValueAsMap = Map.<String, Object>of("additionalPropertName", "additionalPropertyNameValue");
			mapper.mapToSubObjects(Map.of(PROPERTY_NAME, propertyValueAsMap));

			verify(mapper).buildSubObject(PROPERTY_NAME, propertyValueAsMap);
		}

		@Test
		void shouldNotCallBuildSubObjectOnNonMapValue() {
			mapper.mapToSubObjects(Map.of(PROPERTY_NAME, PROPERTY_NAME_VALUE));

			verify(mapper, never()).buildSubObject(any(), any());
		}
	}

	@DisplayName("Build subObject")
	@Nested
	class TestBuildSubObject {

		@Test
		void shouldCallMapAllProperties() {
			mapper.buildSubObject(PROPERTY_NAME, GrpcPropertyTestFactory.createAsMap());

			verify(mapper).mapToAllProperties(GrpcPropertyTestFactory.createAsMap());
		}

		@Test
		void shouldCallMapSubObject() {
			mapper.buildSubObject(PROPERTY_NAME, GrpcPropertyTestFactory.createAsMap());

			verify(mapper).mapToSubObjects(GrpcPropertyTestFactory.createAsMap());
		}

		@Test
		void shouldAddPropertiesToSubObject() {
			var wantedProperty = GrpcPropertyTestFactory.create();

			var grpcSubObject = mapper.buildSubObject(GrpcSubObjectTestFactory.SUB_OBJECT_NAME, Map.of(PROPERTY_NAME, PROPERTY_NAME_VALUE));

			assertThat(grpcSubObject.getPropertyList()).hasSize(1).first().isEqualTo(wantedProperty);
		}
	}

	@Nested
	class TestMapFromGrpcProperty {

		@Test
		@DisplayName("should map grpc property into map")
		void shouldMapProperty() {
			var targetMap = mapper.mapProperty(GrpcPropertyTestFactory.create());

			assertThat(targetMap).isEqualTo(Pair.of(PROPERTY_NAME, PROPERTY_NAME_VALUE));
		}

		@Test
		@DisplayName("should map property with collection as value to map")
		void shouldMapAllPropertyValues() {
			var anotherValue = "anotherPropertyValue";
			var grpcProperty = GrpcPropertyTestFactory.createBuilder().addValue(anotherValue).build();

			var targetMap = mapper.mapProperty(grpcProperty);

			assertThat(targetMap).isEqualTo(Pair.of(PROPERTY_NAME, List.of(PROPERTY_NAME_VALUE, anotherValue)));
		}

		@Test
		@DisplayName("should map property without value to map")
		void shouldMapEmptyProperty() {
			var targetMap = mapper.mapProperty(GrpcPropertyTestFactory.createWithoutValue());

			assertThat(targetMap).isEqualTo(Pair.of(PROPERTY_NAME, null));
		}

		@Test
		void shouldTrimToNull() {
			GrpcProperty grpcProperty = GrpcPropertyTestFactory.createBuilder().addValue(StringUtils.EMPTY).build();

			var resultEntry = mapper.mapProperty(grpcProperty);

			assertThat(resultEntry).isEqualTo(Pair.of(PROPERTY_NAME, Arrays.asList(PROPERTY_NAME_VALUE, null)));
		}

	}

	@Nested
	class TestMapFromGrpcObject {

		@Test
		void shouldMapFromGrpcObject() {
			Map<String, Object> wantedMap = Map.of(PROPERTY_NAME, PROPERTY_NAME_VALUE, SUB_OBJECT_NAME, createSubObjectMap());
			var map = mapper.mapFromGrpc(GrpcObjectTestFactory.create());

			assertThat(map).containsAllEntriesOf(wantedMap);
		}
	}

	@Nested
	class TestMapFromSubObject {

		@Test
		@DisplayName("should map subObject into a map")
		void shouldMapSubObject() {
			var resultMap = mapper.mapSubObject(GrpcSubObjectTestFactory.create());

			assertThat(resultMap).containsAllEntriesOf(createSubObjectMap());
		}

		@Test
		@DisplayName("should map empty property values of subObject to nulls")
		void shouldMapSubObjectEmptyValues() {
			var grpcSubObject = GrpcSubObject.newBuilder().setName(SUB_OBJECT_NAME).addProperty(createWithoutValue()).build();

			var resultMap = mapper.mapSubObject(grpcSubObject);

			assertThat(resultMap).containsEntry(PROPERTY_NAME, null);
		}

	}

	private static Map<String, Object> createSubObjectMap() {
		return Map.of(PROPERTY_NAME, PROPERTY_NAME_VALUE, SUB_SUB_OBJECT_NAME, Map.of(PROPERTY_NAME, PROPERTY_NAME_VALUE));
	}
}
